module Main where

import Control.Arrow
import Control.Applicative
import Control.Applicative.Error
import Control.Concurrent
import Control.Monad.State
import HAppS.Server hiding (Web)
import HAppS.Template.HSP
import HAppS.Template.HSP.Handle
import HSP
import Text.Formlets
import HSP.Formlets
import Index
import Interface

-- * Move to SimpleHTTP

instance Functor Result where
    fmap f NoHandle = NoHandle
    fmap f (Ok g a) = Ok g (f a)
    fmap f (Escape response) = Escape response

instance (Functor m) => Functor (WebT m) where
    fmap f (WebT m) = WebT (fmap (fmap f) m)

instance (Functor m, Monad m) => Applicative (WebT m) where
    pure = return
    (<*>) = ap

-- * Main

main :: IO ()
main =
    do store <- newStore
       tid <- forkIO $ simpleHTTP validateConf $ impl store
       putStrLn "running..."
       waitForTermination
       killThread tid
       destroyStore store

impl :: Store -> [ServerPart Response]
impl store = 
    [ runHSPHandle "." "objfiles" store $ multi $
      withForm "/" "date" dateFull "Index.hs" (ok . toResponse . show)
    ]

-- PROBLEMS
-- 1. there is no type-checking to ensure that pageFP and frm are the same type
-- 2. the params we pass in might collide with other params if there is more than one form
-- 3. supporting more than one form per page period
-- 
-- This should perhaps be split in two. If we have a flash interface,
-- then perhaps we want the POST part to be common? What happens on an
-- error then?
--
-- Perhaps the biggest issue is that Main import Index because it
-- needs the result of createForm + dateFull, which means dynamic
-- reloading is wacked out if you change dateFull.
withForm :: String 
         -> String
         -> Form [Web XML] (WebT (StateT HSPState IO)) a
         -> FilePath
         -> (a -> WebT (StateT HSPState IO) Response)
         -> [ServerPartT (StateT HSPState IO) Response]
withForm action prefix frm pageFP handleOk = 
      [ method GET $  do addParam "formData" ([] :: [(String, String)])
                         addParam "prefix" prefix
                         addParam "action" action
                         addParam "faults" ([] :: [String])
                         execTemplate Nothing pageFP
      , withDataFn lookPairs $ \env ->
        [ method POST $ 
                 do let (extractor, _xml, _ct) = runFormState [] prefix frm
                    res <- extractor (map (second Left) env)
                    case res of
                      Failure faults -> 
                          do addParam "formData" env
                             addParam "prefix" prefix
                             addParam "action" "/"
                             addParam "faults" faults
                             execTemplate Nothing pageFP
                      Success s -> handleOk s
        ]
      ]

