{-# FlexibleInstances, MultiParamTypeClasses, FlexibleContexts, GeneralizedNewtypeDeriving, ScopedTypeVariables, TypeFamilies, TypeSynonymInstances, UndecidableInstances #-}
{-# OPTIONS_GHC -fglasgow-exts -fallow-undecidable-instances -fallow-incoherent-instances -fwarn-missing-signatures #-}
module Generics where

import Control.Applicative (Applicative, (<$>), (<*>))
import Control.Applicative.Error (Failing(..))
import Control.Arrow (second)
import Control.Monad.Identity (Identity)
import Control.Monad.Trans (lift)
import Data.Generics.SYB.WithClass.Basics
import Data.Generics.SYB.WithClass.Instances ()
import HAppS.Data (Default(..), DefaultD(..), defaultProxy)
import HAppS.Server
import HSP
import HSP.Formlets -- (FormHSXT(..), input)
import qualified HSX.XMLGenerator as HSX
import Text.Formlets
import Interface

-- | Given a value of an algebraic type, return a list of the
-- defaultValue for each of the type's constructors.  This should
-- move into HAppS.Extra, or HAppS.Data, or Data.Generics.SYB.
allDefaultValues :: (Data DefaultD a,Default a) => a -> [a]
allDefaultValues x = res
    where res = case datarep $ dataTypeOf defaultProxy x of
                    AlgRep cs ->
                        map (fromConstrB defaultProxy (defaultValueD dict)) cs
                    r ->
                        error ("allDefaultValues: Bad DataRep: " ++ show r)

class FormletOf a where
    formletOf :: (HSX.XMLGenerator x, Monad v, Applicative v) => a -> FormHSXT x v a

data FormletOfD a =
    FormletOfD {formletOfD :: (HSX.XMLGenerator x, Monad v, Applicative v) => a -> FormHSXT x v a}

formletOfProxy :: Proxy FormletOfD
formletOfProxy = error "urk"

instance FormletOf t => Sat (FormletOfD t) where
    dict = FormletOfD { formletOfD = formletOf }

-- |Convert a value of an arbitrary type to a formlet.
instance (Show a, Default a, Ord a, Data FormletOfD a) => FormletOf a where
    formletOf x =
        case dataTypeRep t of
          AlgRep [constr] -> formletOfRecord constr x
          AlgRep constrs -> formletOfUnion constrs x
          other -> error $ "Missing FormletOf instance for type: " ++ show other
        where
          t = dataTypeOf formletOfProxy x

formletOfUnion :: forall a. forall x. forall v. (Show a, Default a, Ord a, Data FormletOfD a, HSX.XMLGenerator x, Monad v, Applicative v) =>
                  [Constr] -> a -> FormHSXT x v a
formletOfUnion constrs x =
    buttons
    where
      buttons = radio pairs (Just (showConstr c)) `check` toValue
      valueForms :: (Show a, Default a, Ord a, Data FormletOfD a, HSX.XMLGenerator x, Monad v, Applicative v) => [FormHSXT x v a]
      valueForms = map formletOf (map snd pairs')
      -- Replace the constructor name with the associated value
      toValue s = maybe (Failure ["Invalid constructor name: " ++ s]) Success (lookup s pairs')
      -- Substitute the particular value passed for the default
      pairs' = map (\ (c', x') -> (showConstr c', if c' == c then x else x')) (zip constrs (allDefaultValues x))
      pairs = map (\ x -> (x, x)) (map showConstr constrs)
      c = toConstr defaultProxy x

formletOfRecord :: forall a. forall x. forall v. (Show a, Default a, Ord a, Data FormletOfD a) => Constr -> a -> FormHSXT x v a
formletOfRecord constr x =
    undefined

instance FormletOf Sample3 where
    formletOf (Constructor6 n s) = 
        Constructor6 <$> inputInteger (Just n) <*> input (Just s)

instance FormletOf Integer where
    formletOf n = inputInteger (Just n)

instance FormletOf String where
    formletOf s = input (Just s)
