{-# LANGUAGE FlexibleInstances, TypeSynonymInstances, MultiParamTypeClasses, FlexibleContexts, TypeFamilies, UndecidableInstances, GeneralizedNewtypeDeriving #-}
{-# OPTIONS_GHC -fno-warn-orphans -F -pgmF trhsx #-}
module Index where

import Data.Generics.SYB.WithClass.Basics
import Control.Applicative
import Control.Applicative.Error
import Control.Monad.Writer
import Data.Monoid
import Control.Monad.Reader
import Control.Monad.State
import Control.Monad.Trans
import Control.Monad.Identity
import HAppS.Data (Default(..))
import HSP
import HSP.Formlets
import Interface
import Text.Formlets
import qualified HSX.XMLGenerator as HSX
import HSP.IdentityT
import HSP.Identity
import Generics

data FormState 
    = FormState { env :: Env
                , prefix :: String
                , action :: String
                , faults :: [String]
                }
      deriving (Read, Show)

createForm :: (HSX.XMLGenerator x, Monad m) => Env -> String -> String -> [String] -> Form [XMLGenT x (HSX.XML x)] m a -> XMLGenT x (HSX.XML x)
createForm env prefix action' faults frm =
    do let (_extractor, xml, _ct) = runFormState env prefix frm
       let faultsXML = if null faults
                       then []
                       else [<p><% show faults %></p>]
       <form action=action' method="post" enctype="multipart/form-data">
        <% faultsXML %>
        <% xml %>
        </form>

-- * demo


dateComponent :: (HSX.XMLGenerator x, Applicative m, Monad m) => Form [XMLGenT x (HSX.XML x)] m Date
dateComponent = Date <$> inputInteger (Just 1) <*> inputInteger (Just 16) <* submit "hit it!"

dateFull ::  (HSX.XMLGenerator x, Applicative m, Monad m) => Form [XMLGenT x (HSX.XML x)] m Date
dateFull = dateComponent `check` ensure validDate "This is not a valid date"

page :: (EmbedAsChild x String, EmbedAsChild x c) => c -> XMLGenT x (HSX.XML x)
page xml =
    <html>
     <head>
      <title>HAppS + HSP + Formlets</title>
     </head>
     <body>
      <% xml %>
     </body>
    </html>

inputFrm :: Form [XMLGenT Identity XML] IO String 
inputFrm = input Nothing <* (submit "go!")

dateForm :: (HSX.XMLGenerator x, Monad v, Applicative v) => 
            Form [XMLGenT x (HSX.XML x)] v Date
dateForm = 
     (\x -> [<p><% x %></p>]) `plug` dateFull

selectForm :: Form [XMLGenT Identity XML] IO String
selectForm = 
    let fruit = ["apple", "banana", "coconut" ] 
    in
      (\x -> [<p><% x %></p>]) `plug` (select (zip fruit fruit) (Just "banana") <* submit "Drink Up!")

sampleForm :: Sample3 -> Form [XMLGenT Identity XML] IO Sample3
sampleForm x =
    -- Put the whole form inside a paragraph
    (\ xml -> [<p><% xml %></p>]) `plug` (formletOf x  <* go)
    where
      -- Put a break in front of the submit button
      go = (\x -> [<span><br/><% x %></span>]) `plug` (submit "Go")
