{-# LANGUAGE FlexibleInstances, FlexibleContexts #-}
{-# OPTIONS_GHC -fno-warn-orphans -F -pgmF trhsx #-}
module Main where

import Control.Arrow
import Control.Applicative
import Control.Applicative.Error
import Control.Concurrent
import Control.Monad.State
import Control.Monad.Reader
import Control.Monad.Identity
import HAppS.Data (defaultValue)
import HAppS.Server hiding (Web)
import HAppS.Server.Extra
import HSP
import HSP.HTML
import Text.Formlets hiding (contentType)
import HSP.Formlets
import Index (page, dateForm, selectForm, sampleForm, createForm)
import Interface
import HSP.IdentityT
import HSP.Identity
import qualified HSX.XMLGenerator as HSX

import qualified Data.ByteString.Char8 as P
import qualified Data.ByteString.Lazy.Char8 as L
import Data.ByteString.Lazy.UTF8 as L
import Generics

-- * Move to SimpleHTTP

instance Functor Result where
    fmap f NoHandle = NoHandle
    fmap f (Ok g a) = Ok g (f a)
    fmap f (Escape response) = Escape response

instance (Functor m) => Functor (WebT m) where
    fmap f (WebT m) = WebT (fmap (fmap f) m)

instance (Functor m, Monad m) => Applicative (WebT m) where
    pure = return
    (<*>) = ap


instance ToMessage (XMLGenT Identity XML) where
    toContentType _ = P.pack "text/html"
    toMessage xml   = toMessage (html4Strict, evalIdentity xml) -- L.fromString (renderAsHTML xml)

instance ToMessage XML where
    toContentType _ = P.pack "text/html"
    toMessage xml   = toMessage (html4Strict, xml) -- L.fromString (renderAsHTML xml)

instance ToMessage (Maybe XMLMetaData, XML) where
    toContentType (Just md,_) = P.pack (contentType md)
    toContentType _ = P.pack "text/html"
    toMessage (Just (XMLMetaData (showDt, dt) _ pr), xml) = 
        L.fromString ((if showDt then (dt ++) else id) (pr xml))
    toMessage (Nothing, xml) =
        L.fromString (renderAsHTML xml)

-- * Main

main :: IO ()
main =
    do tid <- forkIO $ simpleHTTP (validateConf {port = 8008}) $ impl
       putStrLn "running..."
       waitForTermination
       killThread tid
       putStrLn "shutting down..."

impl :: [ServerPart Response]
impl =  [ dir "date"   $ handleForm "prefix" "/date"   page dateForm   (ok . toResponse . show)
        , dir "select" $ handleForm "prefix" "/select" page selectForm (ok . toResponse)
        , dir "sample" $ handleForm "prefix" "/sample" page (sampleForm (defaultValue {int5=8, string5="my string"})) (ok . toResponse . show)
        , method GET $ 
                 ok (toResponse 
                     (evalIdentity 
                      (<html>
                        <head>
                         <title>Index of HAppS+HSP+Formlets examples</title>
                        </head>
                        <body>
                         <p><a href="/date">Date Widget</a></p>
                         <p><a href="/select">Select Widget</a></p>
                         <p><a href="/sample">Sample Widget</a></p>
                        </body>
                       </html>)))
        , debug404
        ]

handleForm :: String -> String -> (XMLGenT Identity XML -> XMLGenT Identity XML) -> Form [XMLGenT Identity XML] IO a -> (a -> WebT IO Response) -> [ServerPart Response]
handleForm prefix action page frm handleOk =
    [ method GET $ ok (toResponse (page (createForm [] prefix action [] frm)))
    , withDataFn lookPairs $ \env' ->
        [ method POST $
                 do let (extractor, _, _) = runFormState [] prefix frm
                        env = map (second Left) env'
                    res <- lift $ extractor env
                    case res of
                      Failure faults ->
                          ok (toResponse (page (createForm env prefix action faults frm)))
                      Success s -> handleOk s
        ]
    ]
